/*
 * COPYRIGHT INFORMATION - DO NOT REMOVE
 * "Portions Copyright (c) 2000-2001 LinuxMagic Inc. All Rights Reserved.
 *
 * This file contains Original Code and/or Modifications of Original Code as
 * defined in and that are subject to the Free Source Code License Version
 * 1.0 (the 'License'). You may not use this file except in compliance with
 * the License. Please obtain a copy of the License at:
 *
 * http://www.linuxmagic.com/opensource/licensing/
 *
 * and read it before using this file.
 *
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND LINUXMAGIC HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT. Please see
 * the License for the specific language governing rights and limitations
 * under the License."
 *
 * Please read the terms of this license carefully. By using or downloading
 * this software or file, you are accepting and agreeing to the terms of this
 * license with LinuxMagic Inc. If you are agreeing to this license on behalf
 * of a company, you represent that you are authorized to bind the company to
 * such a license. If you do not meet this criterion or you do not agree to
 * any of the terms of this license, do NOT download, distribute, use or alter
 * this software or file in any way.
 *
 * Author: Burton Samograd (burton@wizard.ca)
 * CVS Id: $Id: auth.c,v 1.20 2002/12/18 17:21:13 burton Exp $
 *
 * DO NOT MODIFY WITHOUT CONSULTING THE LICENSE
 */

/* C includes */
#define _XOPEN_SOURCE /* for strncmp */
#include <string.h>
#include <unistd.h>

#include "liblm.h"
#include "magic-smtpd.h"
#include "smtp.h"
#include "auth.h"
#include "config.h"
#include "db.h"

static int msd_auth_init_pgdb(void);
static int msd_auth_cleanup_pgdb(void);
static int msd_do_auth_pgdb(int type, lm_string_t username);

/* Can't get rid of warning so include this */
extern int snprintf (char *__restrict __s, size_t __maxlen,
                     __const char *__restrict __format, ...);
int b64decode(lm_string_t instr, lm_string_t* poutstr);
int b64encode(lm_string_t in, lm_string_t* out);

int msd_auth_init(void)
{
  if(config.compat_type != MSD_COMPAT_QMAIL)
    return msd_auth_init_pgdb();
  else
    return -1;
}

int msd_auth_cleanup(void)
{
  if(config.compat_type != MSD_COMPAT_QMAIL)
	return msd_auth_cleanup_pgdb();
  else
    return -1;
}

int msd_do_auth(int type, lm_string_t username64)
{
  if(config.compat_type != MSD_COMPAT_QMAIL)
	return msd_do_auth_pgdb(type, username64);
  else
    return -1;
}

/* Postgres specific authorization functions */
int msd_auth_init_pgdb(void)
{

	int retval = 0;
	/* Nothing to do here, handled by db.c */
	LM_ASSERT(db != NULL);
	if(db == NULL)
	  {
	    /* Can't do auth when db is down */
	    return -1;
	  }

	return retval;
}

int msd_auth_cleanup_pgdb(void)
{
	int retval = 0;
	/* Nothing to do here, handled by db.c */
	LM_ASSERT(db != NULL);
	if(db == NULL)
	  {
	    /* Can't do auth when db is down */
	    return -1;
	  }
	return retval;
}

/* Connect to database and authorize user */
int msd_do_auth_pgdb(int type, lm_string_t username64)
{
	char buffer[1024];
	lm_string_t domain = NULL;
	lm_string_t username = NULL;
	lm_string_t password = NULL;
	lm_string_t password64 = NULL; 
	lm_string_t query = NULL;
	int index;
	int retval = 0;
	PGresult *result;
	
	if(db == NULL)
	  {
	    /* Can't do auth when db is down */
	    return LM_AUTH_ERROR_DB;
	  }

	/* Only support LOGIN auth type for now */
	if(type != SMTP_AUTH_LOGIN)
	{
		LM_LOG_WARN(MSD_LOG_SYS, "Unsupported authorization type.");
		return LM_AUTH_UNSUPPORTED;
	}

	/* Print the Username: line in base64, read an input line and
	   decode it to get the username */
	if(username64 == NULL)
	  {
	    printf("334 VXNlcm5hbWU6\r\n"); /* Username: */
	    fflush(stdout);
	    msd_timeout_read(0, &username64);
	    LM_STRING_CHOP(username64);
	    /* check for user cancel '*' */
	    if(LM_STRING_LEN(username64) == 1 &&
	       LM_STRING_BUFFER(username64)[0] == '*')
	      {
		/* user canceled */
		LM_STRING_FREE(username64);
		return LM_AUTH_CANCEL;
	      }
	  }
	if(b64decode(username64, &username))
	  {
	    LM_ASSERT(username == NULL);
	    LM_STRING_FREE(username64);
	    return LM_AUTH_BAD_DATA;
	  }
	else
	  {
	    LM_STRING_CHOP(username);
	  }
	LM_STRING_FREE(username64);

	/* Print the Password: line in base64, read an input line and
	decode it to get the password */
	printf("334 UGFzc3dvcmQ6\r\n"); /* Password: */
	fflush(stdout);
	msd_timeout_read(0, &password64);
	LM_STRING_CHOP(password64);
	/* check for user cancel '*' */
	if(LM_STRING_LEN(password64) == 1 &&
	   LM_STRING_BUFFER(password64)[0] == '*')
	  {
	    /* user canceled */
	    LM_STRING_FREE(username);
	    LM_STRING_FREE(password64);
	    return LM_AUTH_CANCEL;
	  }
	if(b64decode(password64, &password))
	  {
	    /* failure to decode */
	    LM_ASSERT(password == NULL);
	    LM_STRING_FREE(username);
	    LM_STRING_FREE(password64);
	    return LM_AUTH_BAD_DATA;
	  }
	else
	  {
	    LM_STRING_CHOP(password);
	  }
	LM_STRING_FREE(password64);

	/* See if the username was user@hostname format, and parse out the hostname */
	if(!LM_STRING_CHR(username, '@', index))
	{
	  if(LM_STRING_DUP(domain, username))
	    msd_error_out_of_memory();
	  LM_STRING_START_SHIFT(domain, index+1);
	  LM_STRING_END_SHIFT(username, LM_STRING_LEN(username)-index);
	}
	else
	{
	  if(LM_STRING_NEW(domain, LM_STRING_BUFFER(config.default_domain)))
	    msd_error_out_of_memory();
	}

	/* Could be put into init of the program, but might not be worth it unless
	 * we know that almost all users will be doing auth
	 */
	if(msd_auth_init())
	{
		LM_LOG_ERROR(MSD_LOG_SYS, "Could not open connection to database for AUTH!");
		return LM_AUTH_ERROR_DB;
	}

	/* Build the SQL query to get the password string */
	memset(buffer, 0, sizeof(buffer));

	/* TODO make this query string a config variable */
	snprintf(buffer, sizeof(buffer),
			"select pw_passwd from vpopmail "
			"where "
			"pw_name='%s' "
			"and "
			"pw_domain='%s'",
			LM_STRING_BUFFER(username),
			LM_STRING_BUFFER(domain));
	if(LM_STRING_NEW(query, buffer))
	  msd_error_out_of_memory();
	LM_LOG_MSG(MSD_LOG_SYS, buffer);

	/* TODO check to make sure the db connection is open */
	result = doquery(LM_STRING_BUFFER(query));
	if(PQresultStatus(result) != PGRES_TUPLES_OK)
	{
		/* Didn't get any results so user is not in the database? */
#ifdef _LMDEBUG
		printf("%s\n", PQresStatus(PQresultStatus(result)));
		printf("%s\n", PQresultErrorMessage(result));
#endif
	}
	else
	{
		int ntuples = PQntuples(result);
		int nfields = PQnfields(result);

		/* Got some results */
		if(ntuples == 1 && nfields == 1)
		{

			lm_string_t retpwd = NULL;
			char *tmppwd;
			char *cryptpwd;

			tmppwd = PQgetvalue(result, 0, 0);

			if(tmppwd != NULL)
			{
				LM_STRING_NEW(retpwd, tmppwd);
			}
			else
			{
				/* Wierd error */
				LM_LOG_ERROR(MSD_LOG_SYS, "Postgres lied about the data it returned!");
				retval = LM_AUTH_ERROR;
			}

			cryptpwd = crypt(LM_STRING_BUFFER(password), 
					 LM_STRING_BUFFER(retpwd));
			if(cryptpwd != NULL && retval == 0)
			{
				if(!strncmp(cryptpwd, LM_STRING_BUFFER(retpwd), LM_STRING_LEN(retpwd)))
				{
				  /* Passwords match */
				  /* Kill the relayclient value and set remoteinfo to the username */
				  if(config.tcprelayclient)
				    LM_STRING_FREE(config.tcprelayclient);
				  LM_STRING_NEW(config.tcprelayclient, "");
				  LM_STRING_FREE(config.tcpremoteinfo);
				  LM_STRING_DUP(config.tcpremoteinfo, username);
				  LM_LOG_MSG(MSD_LOG_SYS, "Auth successful");
#ifdef MSD_USE_SYSLOG
				  syslog(LOG_INFO, "AUTH success: %s@%s", 
					 LM_STRING_BUFFER(username), LM_STRING_BUFFER(domain));
#endif
				  retval = 0;
				}
				else
				{
					/* Passwords don't match */
					LM_LOG_MSG(MSD_LOG_SYS, "Auth failed, invalid password.");
#ifdef MSD_USE_SYSLOG
					syslog(LOG_INFO, "AUTH failure: %s@%s", 
							LM_STRING_BUFFER(username), LM_STRING_BUFFER(domain));
#endif
					retval = LM_AUTH_FAILED;
				}
			}
			LM_STRING_FREE(retpwd);
		}
		else
		{
			/* User not in database? */
			LM_LOG_MSG(MSD_LOG_SYS, "Auth failed, user not in database.");
#ifdef MSD_USE_SYSLOG
					syslog(LOG_INFO, "AUTH failure user not in db: %s@%s", 
							LM_STRING_BUFFER(username), LM_STRING_BUFFER(domain));
#endif
			retval = LM_AUTH_NO_USER;
		}
	}

	LM_STRING_FREE(username);
	LM_STRING_FREE(domain);
	LM_STRING_FREE(password);
	LM_STRING_FREE(query);

	return retval;
}
